# -----------------------------------------------------------------------------
#
#  Copyright (c) 2019, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
# -----------------------------------------------------------------------------

import importlib
import unittest

import six


def optional_import(name):
    """
    Optionally import a module, returning None if that module is unavailable.

    Parameters
    ----------
    name : Str
        The name of the module being imported.

    Returns
    -------
    None or module
        None if the module is not available, and the module otherwise.

    """
    try:
        module = importlib.import_module(name)
    except ImportError:
        return None
    else:
        return module


# Commonly-used unittest skip decorators.
cython = optional_import("cython")
requires_cython = unittest.skipIf(cython is None, "Cython not available")

numpy = optional_import("numpy")
requires_numpy = unittest.skipIf(numpy is None, "NumPy not available")

sphinx = optional_import("sphinx")
requires_sphinx = unittest.skipIf(sphinx is None, "Sphinx not available")

traitsui = optional_import("traitsui")
requires_traitsui = unittest.skipIf(traitsui is None, "TraitsUI not available")

requires_python2 = unittest.skipUnless(six.PY2, "Applicable only to Python 2")
conflicts_with_python2 = unittest.skipIf(six.PY2, "Does not apply to Python 2")



# ------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author:        David C. Morrill
#  Original Date: 04/01/2008
#
# ------------------------------------------------------------------------------

""" Defines the TraitValue class, used for creating special, dynamic trait
    values.
"""

# -------------------------------------------------------------------------------
#  Imports:
# -------------------------------------------------------------------------------

from __future__ import absolute_import

from .trait_base import Undefined
from .traits import CTrait
from .has_traits import HasTraits, HasPrivateTraits
from .trait_errors import TraitError
from .trait_types import Tuple, Dict, Any, Str, Instance, Event, Callable
from .trait_handlers import TraitType, _read_only, _write_only, _arg_count

# -------------------------------------------------------------------------------
#  'BaseTraitValue' class:
# -------------------------------------------------------------------------------


class BaseTraitValue(HasPrivateTraits):

    # Subclasses can define this trait as a property:
    # value = Property

    # -- Public Methods ---------------------------------------------------------

    def as_ctrait(self, original_trait):
        """ Returns the low-level C-based trait for this TraitValue.
        """
        notifiers = original_trait._notifiers(0)

        if self._ctrait is not None:
            if (notifiers is None) or (len(notifiers) == 0):
                return self._ctrait

            trait = CTrait(0)
            trait.clone(self._ctrait)
        else:
            trait = self._as_ctrait(original_trait)

        if (
            (trait is not None)
            and (notifiers is not None)
            and (len(notifiers) > 0)
        ):
            trait._notifiers(1).extend(notifiers)

        return trait

    # -- Private Methods --------------------------------------------------------

    def _as_ctrait(self, original_trait):
        """ Returns the low-level C-based trait for this TraitValue.
        """
        value_trait = self.trait("value")
        if value_trait is None:
            return None

        if value_trait.type != "property":
            raise TraitError("Invalid TraitValue specified.")

        metadata = {"type": "property", "_trait_value": self}

        getter, setter, validate = value_trait.property()
        read_only = getter is _read_only
        if not read_only:
            getter = self._getter
            metadata["transient"] = True

        if setter is not _write_only:
            if read_only:
                setter = self._read_only_setter
            else:
                setter = self._setter
            metadata["transient"] = True

        return self._property_trait(getter, setter, validate, metadata)

    def _property_trait(self, getter, setter, validate, metadata):
        """ Returns a properly constructed 'property' trait.
        """
        n = 0
        if validate is not None:
            n = _arg_count(validate)

        trait = CTrait(4)
        trait.property(
            getter, _arg_count(getter), setter, _arg_count(setter), validate, n
        )

        trait.value_allowed(True)
        trait.value_property(True)
        trait.__dict__ = metadata

        return trait

    def _getter(self, object, name):
        return self.value

    def _setter(self, object, name, value):
        old_value = self.value
        self.value = value
        new_value = self.value
        if new_value != old_value:
            object.trait_property_changed(name, old_value, new_value)

    def _read_only_setter(self, object, name, value):
        self.value = value
        object.trait_property_changed(name, Undefined, value)


# -------------------------------------------------------------------------------
#  'TraitValue' class:
# -------------------------------------------------------------------------------


class TraitValue(BaseTraitValue):

    #: The callable used to define a default value:
    default = Callable

    #: The positional arguments to pass to the callable default value:
    args = Tuple

    #: The keyword arguments to pass to the callable default value:
    kw = Dict

    #: The trait to use as the new trait type:
    type = Any

    #: The object to delegate the new value to:
    delegate = Instance(HasTraits)

    #: The name of the trait on the delegate object to get the new value from:
    name = Str

    # -- Private Methods --------------------------------------------------------

    def _as_ctrait(self, original_trait):
        """ Returns the low-level C-based trait for this TraitValue.
        """
        if self.default is not None:
            trait = CTrait(0)
            trait.clone(original_trait)
            if original_trait.__dict__ is not None:
                trait.__dict__ = original_trait.__dict__.copy()

            trait.default_value(7, (self.default, self.args, self.kw))

        elif self.type is not None:
            type = self.type
            try:
                rc = issubclass(type, TraitType)
            except:
                rc = False

            if rc:
                type = type(*self.args, **self.kw)

            if not isinstance(type, TraitType):
                raise TraitError(
                    (
                        "The 'type' attribute of a TraitValue "
                        "instance must be a TraitType instance or subclass, but a "
                        "value of %s was specified."
                    )
                    % self.trait
                )

            self._ctrait = trait = type.as_ctrait()
            trait.value_allowed(True)

        elif self.delegate is None:
            return None

        else:
            if self.name == "":
                raise TraitError(
                    "You must specify a non-empty string "
                    "value for the 'name' attribute when using the "
                    "'delegate' trait of a TraitValue instance."
                )

            metadata = {
                "type": "property",
                "_trait_value": self,
                "transient": True,
            }

            getter = self._delegate_getter
            setter = self._delegate_setter
            validate = None

            self.add_trait("value", Event())
            self.delegate.on_trait_change(self._delegate_modified, self.name)

            trait = self._property_trait(getter, setter, validate, metadata)

        return trait

    def _delegate_getter(self, object, name):
        return getattr(self.delegate, self.name)

    def _delegate_setter(self, object, name, value):
        setattr(self.delegate, self.name, value)

    # -- Traits Event Handlers --------------------------------------------------

    def _delegate_modified(self):
        self.value = True


# -- Helper Function Definitions ------------------------------------------------


def SyncValue(delegate, name):
    return TraitValue(delegate=delegate, name=name)


def TypeValue(type):
    return TraitValue(type=type)


def DefaultValue(default, args=(), kw={}):
    return TraitValue(default=default, args=args, kw=kw)


# -------------------------------------------------------------------------------
#  Tell the C-based traits module about the 'BaseTraitValue' class:
# -------------------------------------------------------------------------------

from . import ctraits

ctraits._value_class(BaseTraitValue)



""" Helper module, providing a common API for tasks that require a different implementation in python 2 and 3.
"""

from __future__ import division, absolute_import

import contextlib

import six

if six.PY2:
    LONG_TYPE = long
else:
    LONG_TYPE = int

# FIXME : These two aliases are being used by released versions of traitsui
# See PR https://github.com/enthought/traitsui/pull/496 which removes their
# use from traitsui.
# Once that PR is merged and a new release of traitsui is available, we can
# remove these aliases for good.
if six.PY2:
    import string

    str_find = string.find
    str_rfind = string.rfind
else:
    str_find = str.find
    str_rfind = str.rfind


if six.PY2:
    from types import InstanceType, ClassType

    def is_old_style_instance(obj):
        return type(obj) is InstanceType

    def is_old_style_class(obj):
        return type(obj) is ClassType

    def is_InstanceType(obj):
        return obj is InstanceType

    def is_ClassType(obj):
        return obj is ClassType


else:

    def is_old_style_instance(obj):
        return False

    def is_old_style_instance(obj):
        return False

    def is_InstanceType(obj):
        return False

    def is_ClassType(obj):
        return False


if six.PY2:
    from types import InstanceType

    def type_w_old_style(obj):
        the_type = type(obj)
        if the_type is InstanceType:
            # Old-style class.
            the_type = obj.__class__
        return the_type
else:
    type_w_old_style = type

if six.PY2:
    from types import ClassType

    ClassTypes = (ClassType, type)
else:
    ClassTypes = (type,)



if six.PY2:
    def nested_context_mgrs(*args):
        return contextlib.nested(*args)
else:
    ExitStack = contextlib.ExitStack

    class nested_context_mgrs(ExitStack):
        """ Emulation of python 2's :py:class:`contextlib.nested`.

        It has gone from python 3 due to it's deprecation status
        in python 2.

        Note that :py:class:`contextlib.nested` was deprecated for
        a reason: It has issues with context managers that fail
        during init. The same caveats also apply here.
        So do not use this unless really necessary!
        """

        def __init__(self, *args):
            super(nested_context_mgrs, self).__init__()
            self._ctxt_mgrs = args

        def __enter__(self):
            ret = []
            try:
                for mgr in self._ctxt_mgrs:
                    ret.append(self.enter_context(mgr))
            except:
                self.close()
                raise
            return tuple(ret)



#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt

from __future__ import absolute_import, print_function, unicode_literals

import unittest

import six

from traits._py2to3 import str_find, str_rfind


class TestPy2to3(unittest.TestCase):
    def test_str_find(self):
        # Note: inputs are Unicode strings.
        self.assertEqual(str_find("abcabc", "c"), 2)
        self.assertEqual(str_find("abcabc", "d"), -1)

    def test_str_rfind(self):
        # Note: inputs are Unicode strings.
        self.assertEqual(str_rfind("abcabc", "c"), 5)
        self.assertEqual(str_rfind("abcabc", "d"), -1)

    if six.PY2:
        def test_str_find_with_bytestrings(self):
            # Try all possible mixes of Unicode with bytes.
            self.assertEqual(str_find("abcabc", "b"), 1)
            self.assertEqual(str_find(u"abcabc", "b"), 1)
            self.assertEqual(str_find("abcabc", u"b"), 1)
            self.assertEqual(str_find(u"abcabc", u"b"), 1)

        def test_str_rfind_with_bytestrings(self):
            # Try all possible mixes of Unicode with bytes.
            self.assertEqual(str_rfind("abcabc", "b"), 4)
            self.assertEqual(str_rfind(u"abcabc", "b"), 4)
            self.assertEqual(str_rfind("abcabc", u"b"), 4)
            self.assertEqual(str_rfind(u"abcabc", u"b"), 4)



# Copyright (c) 2008-2019 by Enthought, Inc.
# All rights reserved.

"""
Version information for this Traits distribution.

This file is autogenerated by the Traits setup.py script.
"""

from __future__ import unicode_literals

#: The full version of the package, including a development suffix
#: for unreleased versions of the package.
version = "5.2.0"

#: The Git revision from which this release was made.
git_revision = "f28c2ae7e32d928bff46532bdd576b8ce74d6c51"



# ------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: David C. Morrill
# Description: <Traits component>
# ------------------------------------------------------------------------------

from __future__ import absolute_import, print_function

import unittest

from traits.api import HasTraits, Property


class Test(HasTraits):

    __traits__ = {}

    def __value_get(self):
        return self.__dict__.get("_value", 0)

    def __value_set(self, value):
        old_value = self.__dict__.get("_value", 0)
        if value != old_value:
            self._value = value
            self.trait_property_changed("value", old_value, value)

    __traits__["value"] = Property(__value_get, __value_set)


class Test_1(Test):
    def value_changed(self, value):
        print("value_changed:", value)


class Test_2(Test):
    def anytrait_changed(self, name, value):
        print("anytrait_changed for %s: %s" % (name, value))


class Test_3(Test_2):
    def value_changed(self, value):
        print("value_changed:", value)


def on_value_changed(value):
    print("on_value_changed:", value)


def on_anyvalue_changed(value):
    print("on_anyvalue_changed:", value)


class TestPropertyNotifications(unittest.TestCase):
    def test_property_notifications(self):
        Test_1().value = "test 1"
        Test_2().value = "test 2"
        Test_3().value = "test 3"

        test_4 = Test()
        test_4.on_trait_change(on_value_changed, "value")
        test_4.value = "test 4"

        test_5 = Test()
        test_5.on_trait_change(on_anyvalue_changed)
        test_5.value = "test 5"

        test_6 = Test()
        test_6.on_trait_change(on_value_changed, "value")
        test_6.on_trait_change(on_anyvalue_changed)
        test_6.value = "test 6"

        test_7 = Test_3()
        test_7.on_trait_change(on_value_changed, "value")
        test_7.on_trait_change(on_anyvalue_changed)
        test_7.value = "test 7"
